
#pragma once

#include "combinator.h"
#include "element_2_string.h"

struct combinator_list
{
   combinator_list()
   {}

   size_t add(combinator const & c)
   {
      size_t t = combinators_.size();
      combinators_.push_back(c);
      return t;
   }

   bool one_step_reduce(element_list & el) const
   {
      if (el.top().is_combinator())
      {
         element c = el.top();
         unsigned t = c.as_combinator();
         el.pop();
         if (!combinators_[t].check_for_apply(el))
         {
            el.push(c);
            return false;
         }

         combinators_[t].apply(el);
         return true;
      }
      else if (el.top().is_ref())
      {
         element_list * nested = el.top().as_ref();
         full_reduce(*nested);

         element hold = el.top();
         el.pop();
         for (element_list::const_iterator i = nested->begin(); i != nested->end(); ++i)
            el.push(*i);

         return true;
      }

      return false;
   }

   void full_reduce(element_list & el) const
   {
      while (one_step_reduce(el));
   }

   void completely_full_reduce(element_list & el)
   {
      full_reduce(el);
      for (element_list::reverse_iterator i = boost::next(el.rbegin()); i != el.rend(); ++i)
      {
         if (i->is_ref())
         {
            element_list * nested = i->as_ref();
            completely_full_reduce(*nested);
            if (nested->size() == 1)
               *i = nested->top();
         }
      }
   }

   combinator operator[](unsigned i)
   {
      return combinators_[i];
   }

   combinator const operator[](unsigned i) const
   {
      return combinators_[i];
   }

private:
   typedef std::vector<combinator> combinators_t;
   combinators_t combinators_;
};
