
#pragma once

namespace lparse
{
   bool is_id(std::wstring const & name)
   {
      for (std::wstring::const_iterator i = name.begin(); i != name.end(); ++i)
         if (!is_alpha_num(*i))
            return false;

      return true;
   }

   lambda_item_ptr exp(preview_tokenizer & inp);
   lambda_item_ptr abs(preview_tokenizer & inp);
   lambda_item_ptr app_list(preview_tokenizer & inp);
   lambda_item_ptr braced(preview_tokenizer & inp);

   lambda_item_ptr match(preview_tokenizer & inp)
   {
      lambda_item_ptr t = exp(inp);
      if (inp.get_next())
         throw std::runtime_error("extra input");

      return t;
   }

   lambda_item_ptr exp(preview_tokenizer & inp)
   {
      boost::optional<std::wstring> t = inp.get_next();
      if (!t)
         throw std::runtime_error("unexpected eof");

      inp.push_back(*t);

      if (*t == L"\\")
         return abs(inp);
      if (is_id(*t) || *t == L"(")
         return app_list(inp);
      else
         throw std::runtime_error("bad input");
   }

   lambda_item_ptr abs(preview_tokenizer & inp)
   {
      boost::optional<std::wstring> t = inp.get_next();
      if (!t)
         throw std::runtime_error("unexpected eof");

      if (*t != L"\\")
         throw std::runtime_error("\\ expected");

      boost::optional<std::wstring> t2 = inp.get_next();
      if (!t2)
         throw std::runtime_error("unexpected eof");

      if (!is_id(*t2))
         throw std::runtime_error("id expected");

      return new abstraction(new variable(*t2), exp(inp));
   }

   lambda_item_ptr app_list(preview_tokenizer & inp)
   {
      boost::optional<std::wstring> t = inp.get_next();
      if (!t)
         throw std::runtime_error("unexpected eof");

      lambda_item_ptr c;

      if (*t == L"(")
      {
         inp.push_back(*t);
         c = braced(inp);
      }
      else if (is_id(*t))
      {
         c = new variable(*t);
      }
      else
         throw std::runtime_error("bad input");

      for (;;)
      {
         boost::optional<std::wstring> t2 = inp.get_next();
         if (!t2)
            return c;

         if (*t2 == L"(")
         {
            inp.push_back(*t2);
            c = new application(c, braced(inp));
         }
         else if (is_id(*t2))
         {
            c = new application(c, new variable(*t2));
         }
         else if (*t2 == L")")
         {
            inp.push_back(*t2);
            return c;
         }
         else
            throw std::runtime_error("bad input");
      }
   }

   lambda_item_ptr braced(preview_tokenizer & inp)
   {
      {
         boost::optional<std::wstring> t = inp.get_next();
         if (!t)
            throw std::runtime_error("unexpected eof");

         if (*t != L"(")
            throw std::runtime_error("( expected");
      }

      lambda_item_ptr r = exp(inp);

      {
         boost::optional<std::wstring> t = inp.get_next();
         if (!t)
            throw std::runtime_error("unexpected eof");

         if (*t != L")")
            throw std::runtime_error(") expected");
      }

      return r;
   }
}
