
#include "stdafx.h"
#include "tokenizer.h"
#include "lambda_item.h"
#include "lambda_parser.h"
#include "lambda_tools.h"
#include "combinator_list.h"
#include "element_2_string.h"
#include "generator.h"
#include "state_machine.h"

struct input_string
{
   typedef boost::optional<wchar_t> result_type;

   input_string(std::wstring const & s)
      : s_(s)
      , p_(0)
   {}

   boost::optional<wchar_t> operator()()
   {
      if (p_ == s_.size())
         return boost::none;
      else
         return s_[p_++];
   }

private:
   std::wstring s_;
   unsigned p_;
};

void do_reductions(combinator_list & cl, element t)
{
   if (!t.is_ref())
      return;

   element_list * el = t.as_ref();
   cl.completely_full_reduce(*el);
   e2s_rec(std::wcout, t);
}

lambda_item_ptr lambda_from_string(std::wstring const & s)
{
   input_string is(s);
   tokenizer::source_t f(is);
   preview_tokenizer pt(f);
   return lparse::match(pt);
}

std::wstring read_file(char const * filename)
{
   FILE * f = fopen(filename, "rb");
   if (!f)
      throw std::runtime_error("fopen failed");

   char c[4096] = {0};
   int t = fread(c, 1, 4096, f);
   fclose(f);

   std::wstring r;
   for (int a = 0; a < t; ++a)
      r += wchar_t(c[a]);

   return r;
}

void eval_trivial(lambda_item_ptr t)
{
   abstraction_elimination_context aec;
   aec.i = new variable(L"I");
   aec.k = new variable(L"K");
   aec.s = new variable(L"S");

   lambda_item_ptr t2 = abstraction_elimination(aec, t);

   t2->print(std::wcout);
   std::wcout << std::endl;

   convert_to_element_context ctec;
   combinator_list cl;
   ctec.i = cl.add(create_i_combinator());
   ctec.k = cl.add(create_k_combinator());
   ctec.s = cl.add(create_s_combinator());

   element e = convert_to_element(aec, ctec, t2);
   //    e2s_rec(std::wcout, e);
   //    std::wcout << std::endl;

   do_reductions(cl, e);
}

void eval_fsm(lambda_item_ptr t)
{
   abstraction_elimination_context aec;
   aec.i = new variable(L"I");
   aec.k = new variable(L"K");
   aec.s = new variable(L"S");

   lambda_item_ptr t2 = abstraction_elimination(aec, t);

   fsm fsm;
   element e = convert_to_element(aec, fsm.get_cte_context(), t2);

//    e2s_rec(std::wcout, e);
//    std::wcout << std::endl;

   if (!e.is_ref())
      return;

   element_list * el = e.as_ref();

   fsm_evaluator fsm_e(fsm);
   fsm_e.completely_full_reduce(*el);
//   e2s_rec(std::wcout, e);

   std::wcout << std::endl << fsm_e.get_reduction_count() << std::endl;
}

int main()
{
//   lambda_item_ptr t = lambda_from_string(L"(\\f \\x f (f x)) 1 2");
//   lambda_item_ptr t = lambda_from_string(L"(\\f \\x f (f x) x x x x x x x) 1 2");
//   lambda_item_ptr t = lambda_from_string(L"(\\a \\b \\f a (b f)) (\\f \\x f (f (f x))) (\\f \\x f (f x)) 1 2");

   lambda_item_ptr t = lambda_from_string(read_file("fact.txt"));
//    t->print(std::wcout);
//    std::wcout << std::endl;
   eval_fsm(t);

   return 0;
}